// Copyright 2001, Jonathan Bachrach.  See file TERMS.

import java.io.*;
import java.util.*;
import antlr.*;

/**
 * List of fragments
 * @author  Jonathan Bachrach
 * @author  Keith Playford
 * @version 0.9, 03OCT00
 */
public class FragmentList extends List {
    public FragmentList () { }
    static public List nil = new FragmentList();
    static { nil.head = nil; nil.tail = nil; }
    public List getNil() { return FragmentList.nil; }
    static public FragmentList FragmentNil = (FragmentList)nil;
    static public FragmentList fnil = (FragmentList)nil;
    public FragmentList getFragmentNil() { return (FragmentList)nil; }
    public Fragment getFragmentHead() { return (Fragment)head; }
    public FragmentList getFragmentListTail() { return (FragmentList)tail; }
    public void setFragmentListTail(FragmentList tl) { tail = tl; }
    public Fragment fhead() { return (Fragment)head; }
    public FragmentList getInsideFragments() { return this; }
    public static FragmentList fcons(Fragment head, FragmentList tail) { 
	return tail.fpush(head); 
    }
    public FragmentList ftail() { return (FragmentList)tail; }
    public Fragment felt(int o) { 
	FragmentList ptr = this;
	for (int i = 0; i == o; i++) {
	    if (i == o)
		return ptr.getFragmentHead();
	    ptr = ptr.getFragmentListTail();
	}
	return null;
    }

    public FragmentList(Fragment hd, FragmentList tl) { head = hd; tail = tl; }
    public FragmentList(FragmentList tl) { tail = tl; }
    static public List make(int sz) {
	FragmentList res = FragmentNil;
	for (int count = 0; count < sz; count++) 
	    res = new FragmentList(res);
	return res;
    }
    static public List make(Fragment[] elts) {
	FragmentList res = FragmentNil;
	for (int i = elts.length - 1; i >= 0; i--) 
	    res = new FragmentList(elts[i], res);
	return res;
    }
    public List cons (Object h, List t) {
	return new FragmentList((Fragment)h, (FragmentList)t);
    }
    public final FragmentList fpush(Fragment x) { 
	if (x instanceof SequenceFragment) {
	    FragmentList ptr = ((SequenceFragment)x).getInsideFragments();
	    FragmentList res = this;
	    for (; ptr != fnil; ptr = ptr.ftail())
		res = res.fpush(ptr.fhead());
	    return res;
	} else {
	    return (FragmentList)(push(x));
	}
    }
    public final FragmentList fpush(int x) {
	return (FragmentList)(push(new IntegerFragment(new Integer(x))));
    }
    public final FragmentList fpush(float x) {
	return (FragmentList)(push(new FloatFragment(new Float(x))));
    }
    public final FragmentList fpush(String x) {
	return (FragmentList)(push(new StringFragment(x)));
    }
    public final FragmentList fpush(char x) {
	return (FragmentList)(push(new CharacterFragment(new Character(x))));
    }
    public static FragmentList flist(Fragment x) {
	return fnil.fpush(x);
    }
    public final FragmentList freverse() {
	return (FragmentList)(reverse());
    }
    public final FragmentList fappend(FragmentList x) {
	return (FragmentList)(append(x));
    }
    public final FragmentList fsublist(int start) {
	return (FragmentList)(sublist(start));
    }
    public final FragmentList fsublist(int start, int end) {
	return (FragmentList)(sublist(start, end));
    }

    public SplitList matchBodyPart () {
	SplitList split = splitAtSemicolon();
	if (split.left == null) 
	    split.left = (FragmentList)nil;
	return split;
    }
    public SplitList matchBodyPartStrict () throws SyntaxMatchFailure {
	SplitList split = splitAtSemicolon();
	if (split.right != null) 
	    return split;
	else 
	    throw new SyntaxMatchFailure();
    }
    public SplitList matchListPart () {
	SplitList split = splitAtComma();
	if (split.left == null) 
	    split.left = (FragmentList)nil;
	return split;
    }
    public SplitList matchListPartStrict () throws SyntaxMatchFailure {
	SplitList split = splitAtComma();
	if (split.right != null) 
	    return split;
	else 
	    throw new SyntaxMatchFailure();
    }
    public FragmentList matchDot () throws SyntaxMatchFailure {
	if (!isEmpty() && getFragmentHead() instanceof DotFragment)
	    return getFragmentListTail();
	else
	    throw new SyntaxMatchFailure();
    }	    
    public FragmentList matchSemicolon () throws SyntaxMatchFailure {
	if (!isEmpty() && getFragmentHead() instanceof SemicolonFragment)
	    return getFragmentListTail();
	else
	    throw new SyntaxMatchFailure();
    }	    
    public FragmentList matchComma () throws SyntaxMatchFailure {
	if (isEmpty() || getFragmentHead() instanceof CommaFragment)
	    return getFragmentListTail();
	else
	    throw new SyntaxMatchFailure();
    }	    
    public FragmentList matchQuestion () throws SyntaxMatchFailure {
	if (!isEmpty() && getFragmentHead() instanceof QuestionFragment)
	    return getFragmentListTail();
	else
	    throw new SyntaxMatchFailure();
    }	    
    public FragmentList matchName (String name) throws SyntaxMatchFailure {
	if (isEmpty()) {
	    throw new SyntaxMatchFailure();
	} else {
	    Fragment f = getFragmentHead();
	    if (!(f instanceof NameFragment)
		|| !((NameFragment)f).getName().equals(name)) 
		throw new SyntaxMatchFailure();
	    else 
		// return new SplitList(flist(f), getFragmentListTail());
		return getFragmentListTail();
	}
    }
    public FragmentList matchName (Symbol name) throws SyntaxMatchFailure {
	return matchName(name.getName());
    }
    public SplitList matchLiteral (Object value) throws SyntaxMatchFailure {
	if (isEmpty()) {
	    throw new SyntaxMatchFailure();
	} else {
	    Fragment f = getFragmentHead();
	    if (!(f instanceof LiteralFragment)
		|| !((LiteralFragment)f).getValue().equals(value))
		throw new SyntaxMatchFailure();
	    else 
		return new SplitList(flist(f), getFragmentListTail());
	}
    }
    public SplitList matchAnyString () throws SyntaxMatchFailure {
	if (isEmpty()) {
	    throw new SyntaxMatchFailure();
	} else {
	    Fragment f = getFragmentHead();
	    if (!(f instanceof StringFragment))
		throw new SyntaxMatchFailure();
	    else 
		return new SplitList(flist(f), getFragmentListTail());
	}
    }
    public SplitList matchExpression () throws SyntaxMatchFailure {
	if (isEmpty()) {
	    throw new SyntaxMatchFailure();
	} else {
	    Fragment f = getFragmentHead();
	    if (!(f instanceof LeafFragment)) // TODO: not quite right
		throw new SyntaxMatchFailure();
	    else 
		return new SplitList(flist(f), getFragmentListTail());
	}
    }
    public SplitList matchTemplate () throws SyntaxMatchFailure {
	if (isEmpty()) {
	    throw new SyntaxMatchFailure();
	} else {
	    Fragment f = getFragmentHead();
	    if (!(f instanceof Template)) 
		throw new SyntaxMatchFailure();
	    else 
		return new SplitList(flist(f), getFragmentListTail());
	}
    }

    public SplitList matchConstraint (SyntaxConstraint c, int size) throws SyntaxMatchFailure {
	LooseFragmentParser.trace("Matching Constraint " + c.getName());
	for (int i = size; i >= 0; i--) {
	    FragmentList fragments = fsublist(0, i);
	    LooseFragmentParser.trace
		("Trying " + c + " at length " + i + " " + fragments);

	    if (c.isAdmissable(fragments)) {
		LooseFragmentParser.trace("  Matched");
		return new SplitList(fragments, fsublist(i));
	    }
	}
	throw new SyntaxMatchFailure(/* "Failed to match constraint " + c.getName() */);
    }
    public SplitList matchConstraint (SyntaxConstraint c) throws SyntaxMatchFailure {
	return matchConstraint(c, size());
    }
    public SplitList matchConstraint (String constraintName, int size) throws SyntaxMatchFailure {
	SyntaxConstraint constraint
	    = (SyntaxConstraint)MacroExpander.lookupConstraint(constraintName);
	return matchConstraint(constraint, size);
    }
    public SplitList matchConstraint (String constraintName) throws SyntaxMatchFailure {
	return matchConstraint(constraintName, size());
    }

    // public boolean matchNameConstraint (SplitList result) {}
    // public boolean matchExpressionConstraint (SplitList result) {}
    // public boolean matchBodyConstraint (SplitList result) {}
    // public boolean matchSymbolConstraint (SplitList result) {}
    public void matchEmpty () throws SyntaxMatchFailure {
	if (!isEmpty())
	    throw new SyntaxMatchFailure();
    }
    public SplitList matchParens () throws SyntaxMatchFailure {
	if (isEmpty()) {
	    throw new SyntaxMatchFailure();
	} else {
	    Fragment f = getFragmentHead();
	    if (!(f instanceof ParensFragment)) {
		throw new SyntaxMatchFailure();
	    } else {
		ParensFragment pf = (ParensFragment)f;
		return new SplitList
		    (pf.getNestedFragments(), getFragmentListTail());
	    }
	}
    }
    public SplitList matchBrackets () throws SyntaxMatchFailure {
	if (isEmpty()) {
	    throw new SyntaxMatchFailure();
	} else {
	    Fragment f = getFragmentHead();
	    if (!(f instanceof BracketsFragment)) {
		throw new SyntaxMatchFailure();
	    } else {
		BracketsFragment nf = (BracketsFragment)f;
		return new SplitList
		    (nf.getNestedFragments(), getFragmentListTail());
	    }
	}
    }
    public SplitList matchBraces () throws SyntaxMatchFailure {
	if (isEmpty()) {
	    throw new SyntaxMatchFailure();
	} else {
	    Fragment f = getFragmentHead();
	    if (!(f instanceof BracesFragment)) {
		throw new SyntaxMatchFailure();
	    } else {
		BracesFragment nf = (BracesFragment)f;
		return new SplitList
		    (nf.getNestedFragments(), getFragmentListTail());
	    }
	}
    }
    // public boolean matchVariable (SplitList result) {}
    // public boolean matchSplicedName (SplitList result) {}
    // static public boolean isMatchingPrefix (String name, String prefix) {}
    // static public boolean isMatchingSuffix (String name, String prefix) {}
    public SplitList splitAtSemicolon () {
	if (isEmpty())
	    return new SplitList(getFragmentNil(), null);
	else if (getFragmentHead() instanceof SemicolonFragment)
	    return new SplitList(getFragmentNil(), getFragmentListTail());
	else {
	    SplitList split = getFragmentListTail().splitAtSemicolon();
	    return new SplitList
		(new FragmentList(getFragmentHead(), split.left), split.right);
	}
    }
    public SplitList splitAtKeptSemicolon () {
	if (isEmpty())
	    return new SplitList(getFragmentNil(), null);
	else if (getFragmentHead() instanceof SemicolonFragment)
	    return new SplitList(getFragmentNil(), this);
	else {
	    SplitList split = getFragmentListTail().splitAtSemicolon();
	    return new SplitList
		(new FragmentList(getFragmentHead(), split.left), split.right);
	}
    }
    public List splitAtSemicolons () {
	List         parts = List.nil;
	FragmentList l     = this;
	while (!l.isEmpty()) {
	    SplitList split = l.splitAtSemicolon();
	    parts = new List(split.left, parts);
	    l     = split.right;
	}
	return parts.reverse();
    }
    public FragmentList skipToComma () {
	if (isEmpty())
	    return getFragmentNil();
	else if (getFragmentHead() instanceof CommaFragment)
	    return getFragmentListTail();
	else
	    return getFragmentListTail().skipToComma();
    }
    public SplitList splitAtComma () {
	if (isEmpty())
	    return new SplitList(getFragmentNil(), null);
	else if (getFragmentHead() instanceof CommaFragment)
	    return new SplitList(getFragmentNil(), getFragmentListTail());
	else {
	    SplitList split = getFragmentListTail().splitAtComma();
	    return new SplitList(new FragmentList(getFragmentHead(), split.left), split.right);
	}
    }
    public List splitAtCommas () {
	List         parts = List.nil;
	FragmentList l     = this;
	while (true) {
	    SplitList split = l.splitAtComma();
	    parts = new List(split.left, parts);
	    if (split.right == null)
		break;
	    l     = split.right;
	}
	return parts.reverse();
    }

    /*
    */
    public FragmentList toPattern () { return null; }
    /// template-substitutions.dylan
    // public Enumeration templateSubstitutions() { }
    /// pattern-elements.dylan
    // public Enumeration bindingMatches() { }
    // public Enumeration boundVariableNames() { }
    // public Enumeration bodyMatchTails() { }

    
    /*
    private class TokenEnum implements Enumeration {
	private Object tokens     = null;
	private List   tokenStack = List.Nil;
	public  TokenEnum(List x) { ptr = x; }
	public  boolean hasMoreElements () {
	    while (true) {
		if (tokens.isEmpty()) {
		    if (tokenStack.isEmpty())
			return false;
		    else {
			tokens     = tokenStack.getHead().getTail();
			tokenStack = tokenStack.getTail();
		    }
		} else if (tokens.getHead() instanceof List) {
		    List val   = (List)tokens.getHead();
		    tokenStack = new List(tokens, tokenStack);
		    tokens     = val;
		} else 
		    return true;
	    }
	}
	public Object nextElement() throws NoSuchElementException {
	    if (hasMoreElements()) {
		Object token = tokens.getHead();
		tokens       = tokens.getTail();
		return token;
	    } else
		throw new NoSuchElementException();
	}
    }
    public Enumeration tokens() {
	return new TokenEnum(this);
    }

    */

    /// PATTERN MATCHING CODE GENERATION

    public Expansion compilePatternElements (IdentifierFragment nowVar, IdentifierFragment endVar) {
	LooseFragmentParser.trace
	    ("compilePatternElements " + this + " : " + nowVar);
	if (isEmpty()) {
	    LooseFragmentParser.trace("empty");
	    if (endVar == null)
		return genMatchEmptyCode(nowVar);
	    else // mark where matching finished for aux rules
		return genSetEndCode(nowVar, endVar);
	} else {
	    return getFragmentHead()
		.compilePatternElement(getFragmentListTail(), nowVar, endVar);
	}
    }

    public Expansion genMatchEmptyCode(IdentifierFragment nowVar) { 
	// #{ ?expression.matchEmpty() } 
	return Template.processTemplate
	    (FragmentList.fnil
	     .fpush(nowVar)
	     .fpush(Template.substituteDot())
	     .fpush(Template.substituteIdentifier("matchEmpty"))
	     .fpush(Template.processParens(FragmentList.fnil))
	     .fpush(Template.substituteSemicolon())
	     .freverse());
    }

    public Expansion genSetEndCode(IdentifierFragment nowVar, IdentifierFragment endVar) { 
	// #{ ?end.setAfter(?expression) } 
	return Template.processTemplate
	    (FragmentList.fnil
	     .fpush(endVar)
	     .fpush(Template.substituteDot())
	     .fpush(Template.substituteIdentifier("setValue"))
	     .fpush(Template.processParens(FragmentList.fnil.fpush(nowVar)))
	     .fpush(Template.substituteSemicolon())
	     .freverse());
    }

    public Fragment compileTemplate(int depth) {
	FragmentList result
	    = fnil
	    .fpush(Template.substituteIdentifier("FragmentList"))
	    .fpush(Template.substituteDot())
	    .fpush(Template.substituteIdentifier("fnil"));
	FragmentList elts = this;
	while (elts != fnil) {
	    Fragment obj = elts.getFragmentHead();
	    Fragment exp = obj.compileTemplate(depth);
	    result = result.fpush(Template.substituteDot())
		.fpush(Template.substituteIdentifier("fpush"))
		.fpush(Template.processParens(FragmentList.fnil.fpush(exp)));
	    elts = elts.getFragmentListTail();
	}
	return Template.processTemplate
	    (result.fpush(Template.substituteDot())
	     .fpush(Template.substituteIdentifier("freverse"))
	     .fpush(Template.processParens(FragmentList.fnil))
	     .freverse());
    }


    List tokens(int limt) {
	// LooseFragmentParser.trace("TOKENS " + limt);
	int         totl = 0;
	List        toks = List.nil;
	Enumeration elts = elements();
	while (elts.hasMoreElements() && totl < limt) {
	    Fragment obj = (Fragment)elts.nextElement();
	    toks = toks.append(obj.tokens());
	    totl = totl + 1;
	}
	return toks;
    }

    List tokens() {
	return tokens(size());
    }

    public FragmentList expandTemplates() {
	FragmentList ptr = this;
	while (ptr != fnil) {
	    ptr.head = ptr.getFragmentHead().expandTemplates();
	    ptr      = ptr.getFragmentListTail();
	}
	return this;
    }

    public FragmentList compileQuestionExpression (int tDepth, int qDepth) {
	FragmentList ptr = this;
	FragmentList res = fnil;
	while (ptr != fnil) {
	    Fragment hd
		= ptr.getFragmentHead();
	    Fragment e
		= hd.compileQuestionExpression(tDepth, qDepth);
	    res = res.fpush(e);
	}
	return res.freverse();
    }

    public FragmentList expandMacros
	  (MacroExpander expander, boolean isRecursive) {
	FragmentList  ptr = this; // next place to consider for expansion
	for (ptr = this; ptr != fnil; ptr = ptr.ftail()) {
	    Fragment elt = ptr.fhead();
	    ptr.head = elt.expandMacros(expander, isRecursive);
        }
	return this;
    }

    public FragmentList markMacros (MacroExpander expander) {
	FragmentList  res = this; // final result
	FragmentList  ptr = this; // next place to consider for expansion
	FragmentList mark = fnil; // beginning of macro
	while (ptr != fnil) {
	    Fragment elt = ptr.getFragmentHead();
	    if (elt instanceof IdentifierFragment) {
		IdentifierFragment id = (IdentifierFragment)elt;
		// System.out.println("CHECKING " + id.getName());
		SyntaxExpander macro  = expander.isMacro(id);
		if (macro != null) {
		    // System.out.println("MACRO " + macro);
		    ExpansionResult er
			= expander.markMacro(id, mark.ftail(), ptr.ftail(), macro);
		    MacroCall call = er.getMacroCall();
		    ptr = er.getTail();  // end of macro
		    LooseFragmentParser.trace("MARK = " + mark);
		    LooseFragmentParser.debugOut("END", ptr);
		    LooseFragmentParser.debugOut("CALL", call.getInsideFragments());
		    FragmentList args = call.getIdAndBeyond(); 
		    LooseFragmentParser.debugOut("ARGS", args);
		    args.tail = (args.ftail().markMacros(expander));
		    FragmentList newTail // insert call before end
			= ptr.fpush(call);
		    LooseFragmentParser.debugOut("NEWTAIL", newTail);
		    if (mark == fnil) {  // at very beginning still
			res = newTail;
		    } else
			mark.setFragmentListTail(newTail);
		    LooseFragmentParser.debugOut("PTR", ptr);
		    mark = newTail;
		    // this.pprint(Fragment.out, 0, true); 
		} else {
		    // progress modifiers cause id
		    ptr  = ptr.ftail();
		}
	    } else {
		ptr.head = elt.markMacros(expander);
		mark     = ptr;          // reset modifiers cause not id
		ptr      = ptr.ftail();
	    }
	}
	return res;
    }

    void print(PrettyStream p) {
	Enumeration elts = elements();
	while (elts.hasMoreElements()) {
	    Fragment obj = (Fragment)elts.nextElement();
	    obj.print(p);
	}
    }
    void println(PrettyStream p) {
	Enumeration elts = elements();
	while (elts.hasMoreElements()) {
	    Fragment obj = (Fragment)elts.nextElement();
	    obj.println(p);
	}
    }
    void pprints(PrettyStream p, int depth, boolean isOuterLast, boolean isCode) {
	Enumeration elts = elements();
	boolean isMore   = elts.hasMoreElements();
	while (isMore) {
	    Fragment obj = (Fragment)elts.nextElement();
	    isMore = elts.hasMoreElements();
	    obj.pprinte(p, depth, !isMore && isOuterLast, isCode);
	}
    }
    void pprint(PrettyStream p) {
	pprint(p, 0, true);
    }
    void pprint(PrettyStream p, int depth, boolean isCode) {
	pprints(p, depth, true, isCode);
    }
    void pprinte(PrettyStream p, int depth, boolean isLast, boolean isCode) {
	pprints(p, depth, isLast, isCode);
    }
    void pprint(PrettyStream p, int depth) { pprint(p, depth, false); }
    void lprints(PrettyStream p, int depth, boolean isOuterLast, boolean isCode) {
	Enumeration elts = elements();
	boolean isMore   = elts.hasMoreElements();
	while (isMore) {
	    Fragment obj = (Fragment)elts.nextElement();
	    isMore = elts.hasMoreElements();
	    obj.lprinte(p, depth, !isMore && isOuterLast, isCode);
	}
    }
    void lprint(PrettyStream p, int depth, boolean isCode) {
	lprints(p, depth, true, isCode);
    }
    void lprint(PrettyStream p) {
	lprint(p, 0, true);
    }
    void lprinte(PrettyStream p, int depth, boolean isLast, boolean isCode) {
	lprints(p, depth, isLast, isCode);
    }
    void lprint(PrettyStream p, int depth) { lprint(p, depth, false); }
}
