// Copyright 2001, Jonathan Bachrach.  See file TERMS.

import java.io.*;
import java.util.*;
import antlr.*;

/**
 * Constraint expressible through nonterminals in Java grammar 
 * @author  Jonathan Bachrach
 * @author  Keith Playford
 * @version 0.9, 03OCT00
 */
public abstract class GrammarSyntaxConstraint extends BaseSyntaxConstraint {
    public abstract String getName();
    public boolean isAdmissable (FragmentList fragments) {
	LooseFragmentParser.trace("Fragments: " + fragments);
	return isAdmissable(fragments.tokens());
    }
    public boolean isAdmissable (List tokens) {
	LooseFragmentParser.trace("Tokens(" + tokens.size() + ") " + tokens);
	
	JavaTokenLexer lexer      = new JavaTokenLexer(tokens);
	JavaRecognizer javaParser = new JavaRecognizer(lexer);

	try {
	    parse(lexer, javaParser);
	    boolean isConsumed = lexer.isEmpty();
	    if (isConsumed)
		LooseFragmentParser.trace
		    ("Succeeded to parse " + getName() + " constraint ");
	    else 
		LooseFragmentParser.trace
		    ("Failed to parse/consume " 
		     + getName() + " constraint " + lexer.nextToken());
	    return isConsumed;
	} catch (Exception e) {
	    e.printStackTrace();
	    LooseFragmentParser.trace
		("Failed to parse " + getName() + " constraint " + e);
	    return false;
	}
    }
    public abstract void parse (JavaTokenLexer lexer, JavaRecognizer parser)
	throws Exception;
}
