// Copyright 2001, Jonathan Bachrach.  See file TERMS.

import java.io.*;
import java.util.*;

/**
 * Good old proper Lisp lists
 * @author  Jonathan Bachrach
 * @author  Keith Playford
 * @version 0.9, 03OCT00
 */
public class List /* implements Cloneable */ {
    public List() {}
    static public List nil = new List();
    static { nil.head = nil; nil.tail = nil; }
    public List getNil() { return nil; }
    public List(Object hd, List tl) { head = hd; tail = tl; }
    static public List list (Object x) {
	return new List(x, nil);
    }
    static public List make(int sz) {
	List res = nil;
	for (int count = 0; count < sz; count++) 
	    res = new List(null, res);
	return res;
    }
    static public List make(Object[] elts) {
	List res = nil;
	for (int i = elts.length - 1; i >= 0; i--) 
	    res = new List(elts[i], res);
	return res;
    }
    Object head = null;
    public Object getHead() { return head; }
    public Object setHead(Object x) { return (head = x); }
    List   tail = null;
    public List getTail() { return tail; }
    public List setTail(List x) { return (tail = x); }
    private class HeadEnum implements Enumeration {
	private List ptr = null;
	public  HeadEnum(List x) { ptr = x; }
	public  boolean hasMoreElements () {
	    return !ptr.isEmpty();
	}
	public Object nextElement() throws NoSuchElementException {
	    if (hasMoreElements()) {
		Object hd = ptr.head;
		ptr = ptr.tail;
		return hd;
	    } else
		throw new NoSuchElementException();
	}
    }
    public Enumeration elements() {
	return new HeadEnum(this);
    }

    private class TailEnum implements Enumeration {
	private List ptr = null;
	public  TailEnum(List x) { ptr = x; }
	public  boolean hasMoreElements () {
	    return !ptr.isEmpty();
	}
	public Object nextElement() throws NoSuchElementException {
	    if (hasMoreElements()) {
		Object tl = ptr;
		ptr = ptr.tail;
		return tl;
	    } else
		throw new NoSuchElementException();
	}
    }
    public Enumeration tails() {
	return new TailEnum(this);
    }
    public int size() { 
	List ptr   = this;
	int  count = 0;
	while (!ptr.isEmpty()) {
	    count += 1;
	    ptr = ptr.tail;
	}
	return count;
    }
    public List cons (Object h, List t) {
	return new List(h, t);
    }
    public List reverse() { 
	List ptr   = this;
	List ans   = getNil();
	while (!ptr.isEmpty()) {
	    ans = cons(ptr.head, ans);
	    ptr = ptr.tail;
	}
	return ans;
    }
    public List append(List x) { 
	if (x.isEmpty())
	    return this;
	else if (isEmpty())
	    return x;
	else {
	    List ptr   = this;
	    List rev   = getNil();
	    while (!ptr.isEmpty()) {
		rev = cons(ptr.head, rev);
		ptr = ptr.tail;
	    }
	    List res = x;
	    while (!rev.isEmpty()) {
		List p = rev; rev = p.tail; p.tail = res;
		res = p;
	    }
	    return res;
	}
    }
    public List sublist(int start) { 
	List l = this;
	for (int i = 0; i < start && l != getNil(); i++) 
	    l = l.tail;
	List r = getNil();
	for (; l != getNil(); l = l.tail) 
	    r = r.push(l.head);
	return r.reverse();
    }
    public List sublist(int start, int end) { 
	int  i = 0;
	List l = this;
	for (; i < start && l != getNil(); i++) 
	    l = l.tail;
	List r = getNil();
	for (int j = i; j < end && l != getNil(); l = l.tail, j++) 
	    r = r.push(l.head);
	return r.reverse();
    }
    public final boolean contains(Object x) {
	List ptr = this;
	while (!ptr.isEmpty()) {
	    if (ptr.head == x)
		return true;
	    ptr = ptr.tail;
	}
	return false;
    }
    public final List push(Object x) {
	return cons(x, this);
    }
    public boolean isEmpty() {
	return (this == getNil());
    }
    public String toString() {
	String  s       = "(";
	List    ptr     = this;
	boolean isFirst = true;
	while (!ptr.isEmpty()) {
	    if (isFirst)
		isFirst = false;
	    else
		s += ", ";
	    if (ptr.head == null) {
		LooseFragmentParser.trace("Null head");
		break;
	    }
	    s += ptr.head.toString();
	    if (ptr.tail == null) {
		LooseFragmentParser.trace("Null tail");
		break;
	    }
	    ptr = ptr.tail;
	}
	return (s + ")");
    }
}
