// Copyright 2001, Jonathan Bachrach.  See file TERMS.

import java.lang.*;
import java.io.*;

/**
 * A class that controls macro expansion
 * @author  Jonathan Bachrach
 * @author  Keith Playford
 * @version 0.9, 03OCT00
 */
public class MacroExpander {
    boolean _haveMacros = false;

    SpecialClassLoader macroLoader
	= new SpecialClassLoader("SExpander");
    private static SpecialClassLoader constraintLoader
	= new SpecialClassLoader("SConstraint");

    public MacroExpander () { }

    public static SyntaxConstraint lookupConstraint (String constraintName) {
	String name
	    = (constraintName.equals("*")) ? "wildcard" : constraintName;
	// try {
	    SyntaxConstraint c
		= (SyntaxConstraint)constraintLoader.loadInstance(name);
	    LooseFragmentParser.trace
		("Looking up constraint " + constraintName + " found " + c);
	    return c;
        // } catch (Exception e) {
        //    throw new SyntaxConstraintFailure();
	// }
    }

    public boolean haveMacros () { return _haveMacros; }
    public void setHaveMacros (boolean macrosp) { _haveMacros = macrosp; }

    public boolean isContinuationWord (IdentifierFragment id, String[] words) {
	for (int i = 0; i < words.length; i++)
	    if (id.getName().equals(words[i]))
		return true;
	return false;
    }

    public SplitList getMacroArguments 
	  (IdentifierFragment id, FragmentList fragments, String[] cwords) {
	FragmentList args = FragmentList.fnil.fpush(id);
	Fragment      arg = null;
	while (true) {
	    do {
		if (fragments.isEmpty())
		    break;
		arg       = fragments.fhead();
		fragments = fragments.ftail();
		args      = args.fpush(arg);
	    } while (!arg.isTerminator());
	    if (fragments.isEmpty())
		break;
	    else {
		FragmentList oFragments = fragments;
		arg       = fragments.fhead();
		fragments = fragments.ftail();
		if (arg instanceof IdentifierFragment 
		    && isContinuationWord
		         ((IdentifierFragment)arg, cwords)) {
		    args      = args.fpush(arg);
		    continue;
		} else {
		    fragments = oFragments;
		    break;
		}
	    }
	}
	args = args.freverse();
	return new SplitList(args, fragments);
    }

    public FragmentList copyModifiers (FragmentList mods, IdentifierFragment id) {
	FragmentList res = FragmentList.fnil;
	
	for (FragmentList ptr = mods; ptr != FragmentList.fnil; ptr = ptr.ftail()) {
	    Fragment x = ptr.fhead();
	    if (x == id)
		return res.freverse();
	    else
		res = res.fpush(x);
	}
	return FragmentList.fnil;
    }

    public ExpansionResult markMacro
	  (IdentifierFragment id, FragmentList modsMark,
	   FragmentList fragments, SyntaxExpander expander) {
	// TODO: could be more efficient to not include identifier in match
	SplitList argumentsResult = null;
	String[] continuationWords
	    = expander.getContinuationWords();
	FragmentList modifiers
	    = copyModifiers(modsMark, id);
	argumentsResult
	    = getMacroArguments(id, fragments, continuationWords);
	FragmentList arguments
	    = argumentsResult.getBefore();

	LooseFragmentParser.debugOut("Macro Continuation Words " + continuationWords);
	LooseFragmentParser.debugOut("Macro Modifiers " + modifiers);
	LooseFragmentParser.debugOut("Macro Arguments", arguments);

	arguments      = modifiers.fappend(arguments);
	setHaveMacros(true);
	MacroCall call = new MacroCall(expander, modifiers, id, arguments);
	call.setCurrentLine(id.getLine());
	return new ExpansionResult(call, argumentsResult.getAfter());
    }

    public SyntaxExpander isMacro (IdentifierFragment id) {
	SyntaxExpander syntaxExpander
	    = (SyntaxExpander)macroLoader.loadInstance(id.getName());
	return syntaxExpander;
    }
}
