// Copyright 2001, Jonathan Bachrach.  See file TERMS.

import java.io.*;
import java.util.*;
import java.lang.reflect.*;

/**
 * A class loader for syntax expanders and constraints
 * @author  Jonathan Bachrach
 * @author  Keith Playford
 * @version 0.9, 03OCT00
 */
public class SpecialClassLoader extends ClassLoader {
    private Map classes = new HashMap();
    private Map instances = new HashMap();
    private Set notClasses = new HashSet();
    private String suffix = "";
    SpecialClassLoader() {
    }
    SpecialClassLoader(String suffix) {
	this.suffix = suffix;
    }

    public boolean notClass(String name) {
	return notClasses.contains(name);
    }

    public synchronized Class loadClass(String name, boolean resolve) {
	String className = name + suffix;
	Class cl = (Class) classes.get(className);

	if (cl == null) {                // i.e., cl is a new class
	    if (notClasses.contains(className)) 
		return null;
	    else {
		// LooseFragmentParser.trace("Looking up " + className);
		try { 
		    Class c = findSystemClass(className); 
		    return c;
		} catch (ClassNotFoundException e) {
		} catch (NoClassDefFoundError e) {
		}

		// load class bytes
		byte[] classBytes = loadClassBytes(className);
		if (classBytes == null) {
		    notClasses.add(className);
		    return null;
		}

		LooseFragmentParser.trace("  Defining " + className);
		cl = defineClass(className, classBytes, 0, classBytes.length);
		if (cl == null) {
		    notClasses.add(className);
		    return null;
		}

		LooseFragmentParser.trace("  Remembering " + className);
		classes.put(className, cl);        // remember the class
	    }
	}

	if (resolve) resolveClass(cl);
	return cl;
    }
    public synchronized Class loadClass(String name) {
	return loadClass(name, false);
    }

    public synchronized Object loadInstance(String name, boolean resolve) {
	Class classof = loadClass(name);
	if (classof == null) {
	    return null;
	} else {
	    Object instance = instances.get(classof);
	    if (instance == null) {
		try {
		    instance = classof.newInstance();
		    instances.put(classof, instance);
		} catch (Exception e) {
		    LooseFragmentParser.trace("  Couldn't create class " + name);
		    return null;
		}
	    }
	    return instance;
	}
    }
    public synchronized Object loadInstance(String name) {
	return loadInstance(name, false);
    }

    private byte[] loadClassBytes(String name) {
	String className = name.replace('.', '/') + ".class";
	FileInputStream in = null;
	try {
	    in = new FileInputStream(className);
	    ByteArrayOutputStream buffer = new ByteArrayOutputStream();
	    int ch;
	    while ((ch = in.read()) != -1) {
		byte b = (byte) ch;
		buffer.write(b);
	    }
	    in.close();
	    return buffer.toByteArray();
	} catch(IOException e) {
	    if (in != null) {
		try { in.close(); }
		catch (IOException e2) {}
	    }
	    return null;
	}
    }
}
